"""Lock to prevent sending multiple command at once to Epson projector."""

import time
from .const import (TURN_ON, TURN_OFF, INV_SOURCES, SOURCE, ALL, TIMEOUT_TIMES)


class Lock:
    def __init__(self):
        """Init lock for sending request to projector when it is busy."""
        self._isLocked = False
        self._timer = 0
        self._operation = False

    def setLock(self, command):
        """Set lock on requests."""
        if command in (TURN_ON, TURN_OFF):
            self._operation = command
        elif command in INV_SOURCES:
            self._operation = SOURCE
        else:
            self._operation = ALL
        self._isLocked = True
        self._timer = time.time()

    def __unlock(self):
        """Unlock sending requests to projector."""
        self._operation = False
        self._timer = 0
        self._isLocked = False

    def checkLock(self):
        """
        Lock checking.

        Check if there is lock pending and check if enough time
        passed so requests can be unlocked.
        """
        if self._isLocked:
            if (time.time() - self._timer) > TIMEOUT_TIMES[self._operation]:
                self.__unlock()
                return False
            return True
        return False



"""HTTP connection of Epson projector module."""
import logging

import aiohttp
import asyncio
import async_timeout

from .const import (ACCEPT_ENCODING, ACCEPT_HEADER, BUSY,
                    EPSON_KEY_COMMANDS, HTTP_OK, STATE_UNAVAILABLE)
from .projector_tcp import ProjectorTcp
_LOGGER = logging.getLogger(__name__)


class ProjectorHttp:
    """
    Epson projector class.

    Control your projector with Python.
    """

    def __init__(self, host, websession, port=80, encryption=False, loop=None):
        """
        Epson Projector controller.

        :param str ip:          IP address of Projector
        :param int port:        Port to connect to. Default 80.
        :param bool encryption: User encryption to connect

        """
        self._host = host
        self._port = port
        self._encryption = encryption
        http_proto = 'https' if self._encryption else 'http'
        self._http_url = '{http_proto}://{host}:{port}/cgi-bin/'.format(
            http_proto=http_proto,
            host=self._host,
            port=self._port)
        referer = "{http_proto}://{host}:{port}/cgi-bin/webconf".format(
            http_proto=http_proto,
            host=self._host,
            port=self._port)
        self._headers = {
            "Accept-Encoding": ACCEPT_ENCODING,
            "Accept": ACCEPT_HEADER,
            "Referer": referer
        }
        self.websession = websession
        self._loop = loop
        self._tcp_projector = None

    def close(self):
        return

    async def get_serial_number(self, timeout):
        if self._loop and self._host:
            if not self._tcp_projector:
                self._tcp_projector = ProjectorTcp(host=self._host, loop=self._loop)
            test = await self._tcp_projector.get_property("EEMP0100À¨E", timeout)
        return STATE_UNAVAILABLE

    async def get_property(self, command, timeout):
        """Get property state from device."""
        response = await self.send_request(
            timeout=timeout,
            params=EPSON_KEY_COMMANDS[command],
            type='json_query')
        if not response:
            return False
        try:
            if response == STATE_UNAVAILABLE:
                return STATE_UNAVAILABLE
            return response['projector']['feature']['reply']
        except KeyError:
            return BUSY

    async def send_command(self, command, timeout):
        """Send command to Epson."""
        response = await self.send_request(
            timeout=timeout,
            params=EPSON_KEY_COMMANDS[command],
            type='directsend',
            command=command)
        return response

    async def send_request(self, params, timeout,
                           type='json_query', command=False):
        """Send request to Epson."""
        try:
            with async_timeout.timeout(timeout):
                url = '{url}{type}'.format(
                    url=self._http_url,
                    type=type)
                async with self.websession.get(
                    url=url, params=params,
                        headers=self._headers) as response:
                    if response.status != HTTP_OK:
                        _LOGGER.warning(
                            "Error message %d from Epson.", response.status)
                        return False
                    if type == 'json_query':
                        return await response.json()
                    return response
        except (aiohttp.ClientError, aiohttp.ClientConnectionError, TimeoutError, asyncio.exceptions.TimeoutError):
            _LOGGER.error("Error request")
            return STATE_UNAVAILABLE



"""Main of Epson projector module."""
import logging

from .const import (BUSY, TCP_PORT, HTTP_PORT)
from .timeout import get_timeout

from .lock import Lock

_LOGGER = logging.getLogger(__name__)


class Projector:
    """
    Epson projector class.

    Control your projector with Python.
    """

    def __init__(self, host, loop, websession=None, type='tcp',
                 encryption=False, timeout_scale=1.0):
        """
        Epson Projector controller.

        :param str host:        Hostname/IP/serial to the projector
        :param obj loop:        Asyncio loop to pass for TCP/serial connection
        :param obj websession:  Websession to pass for HTTP protocol
        :param bool encryption: User encryption to connect, only for HTTP.
        :param timeout_scale    Factor to multiply default timeouts by (for slow projectors)

        """
        self._lock = Lock()
        self._type = type
        self._timeout_scale = timeout_scale
        if self._type == 'http':
            self._host = host
            from .projector_http import ProjectorHttp
            self._projector = ProjectorHttp(host, websession,
                                            HTTP_PORT, encryption, loop)
        elif self._type == 'tcp':
            from .projector_tcp import ProjectorTcp
            self._host = host
            self._projector = ProjectorTcp(host, TCP_PORT, loop)
        elif self._type == 'serial':
            from .projector_serial import ProjectorSerial
            self._host = host
            self._projector = ProjectorSerial(host, loop)

    def close(self):
        """Close connection. Not used in HTTP"""
        self._projector.close()

    def set_timeout_scale(self, timeout_scale=1.0):
        self._timeout_scale = timeout_scale

    async def get_serial_number(self):
        return await self._projector.get_serial()

    async def get_property(self, command, timeout=None):
        """Get property state from device."""
        _LOGGER.debug("Getting property %s", command)
        timeout = timeout if timeout else get_timeout(command, self._timeout_scale)
        if self._lock.checkLock():
            return BUSY
        return await self._projector.get_property(command,
                                                  timeout)

    async def send_command(self, command):
        """Send command to Epson."""
        _LOGGER.debug("Sending command to projector %s", command)
        if self._lock.checkLock():
            return False
        self._lock.setLock(command)
        return await self._projector.send_command(command,
                                                  get_timeout(command, self._timeout_scale))

    async def send_request(self, command):
        """Get property state from device."""
        _LOGGER.debug("Getting property %s", command)
        if self._lock.checkLock():
            return BUSY
        return await self._projector.send_request(params=command, timeout=10)




"""TCP connection of Epson projector module."""
import logging

import asyncio
import async_timeout

from .const import (
    BUSY,
    ESCVPNET_HELLO_COMMAND,
    ESCVPNETNAME,
    ERROR,
    CR,
    CR_COLON,
    GET_CR,
    EPSON_CODES,
    POWER,
    PWR_ON,
    SERIAL_BYTE,
    TCP_SERIAL_PORT,
)
from .timeout import get_timeout

_LOGGER = logging.getLogger(__name__)


class ProjectorTcp:
    """
    Epson TCP connector
    """

    def __init__(self, host, port=3629, loop=None):
        """
        Epson TCP connector

        :param str host:    IP address of Projector
        :param int port:    Port to connect to. Default 3629.
        :param obj loop:    Pass asyncio loop to connector

        """
        self._host = host
        self._port = port
        self._isOpen = False
        self._serial = None
        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop

    async def async_init(self):
        """Async init to open connection with projector."""
        returnvalue = False
        try:
            with async_timeout.timeout(10):
                self._reader, self._writer = await asyncio.open_connection(
                    host=self._host, port=self._port, loop=self._loop
                )
                self._writer.write(ESCVPNET_HELLO_COMMAND.encode())
                response = await self._reader.read(16)
                if response[0:10].decode() == ESCVPNETNAME and response[14] == 32:
                    self._isOpen = True
                    _LOGGER.info("Connection open")
                    returnvalue = True
                else:
                    _LOGGER.info("Cannot open connection to Epson")
                    returnvalue = False
        except asyncio.TimeoutError:
            _LOGGER.error("Timeout error")
            returnvalue = False
        except ConnectionRefusedError:
            _LOGGER.error("Connection refused Error")
            returnvalue = False
        return returnvalue

    def close(self):
        if self._isOpen:
            self._writer.close()

    async def get_property(self, command, timeout):
        """Get property state from device."""
        response = await self.send_request(timeout=timeout, command=command + GET_CR)
        if not response:
            return False
        try:
            return response.split("=")[1]
        except KeyError:
            return BUSY

    async def send_command(self, command, timeout):
        """Send command to Epson."""
        response = await self.send_request(timeout=timeout, command=command + CR)
        return response

    async def send_request(self, timeout, command):
        """Send TCP request to Epson."""
        if self._isOpen is False:
            await self.async_init()
        if self._isOpen and command:
            with async_timeout.timeout(timeout):
                self._writer.write(command.encode())
                response = await self._reader.read(16)
                response = response.decode().replace(CR_COLON, "")
                if response == ERROR:
                    return False
                return response

    async def get_serial(self):
        """Send TCP request for serial to Epson."""
        if not self._serial:
            if self._isOpen is False:
                await self.async_init()
            try:
                with async_timeout.timeout(10):
                    power_on = await self.get_property(POWER, get_timeout(POWER))
                    if power_on == EPSON_CODES[POWER]:
                        reader, writer = await asyncio.open_connection(
                            host=self._host, port=TCP_SERIAL_PORT, loop=self._loop
                        )
                        _LOGGER.debug("Asking for serial number.")
                        writer.write(SERIAL_BYTE)
                        response = await reader.read(32)
                        self._serial = response[24:].decode()
                        writer.close()
                    else:
                        _LOGGER.error(
                            "Is projector turned on?"
                        )
            except asyncio.TimeoutError:
                _LOGGER.error(
                    "Timeout error receiving SERIAL of projector. Is projector turned on?"
                )
        return self._serial



from .const import TIMEOUT_TIMES

def get_timeout(self, command=None, timeout_scale=1):
        if command in TIMEOUT_TIMES:
            return TIMEOUT_TIMES[command] * timeout_scale
        else:
            return TIMEOUT_TIMES['ALL'] * timeout_scale


"""Serial USB/RS232 connection of Epson projector module."""
import logging

import asyncio
import serial_asyncio
from serial.serialutil import SerialException
from .const import (ESCVP_HELLO_COMMAND, COLON, CR, GET_CR,
                    BUSY, ERROR)
import async_timeout

_LOGGER = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10


class ProjectorSerial:
    """
    Epson Serial connector
    """

    def __init__(self, host, loop=None):
        """
        Epson Serial connector

        :param str host:    Device to connect to.
        :param obj loop:    Pass asyncio loop to connector

        """
        self._host = host
        self._reader = None
        self._writer = None
        self._isOpen = False
        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop

    async def async_init(self):
        """Async init to open serial connection with projector."""
        try:
            with async_timeout.timeout(DEFAULT_TIMEOUT):
                (self._reader,
                 self._writer) = await serial_asyncio.open_serial_connection(
                    url=self._host,
                    baudrate=9600,
                    loop=self._loop)
                if (self._reader and self._writer):
                    self._isOpen = True
                    self._writer.write(ESCVP_HELLO_COMMAND.encode())
                    response = await self._reader.readuntil(COLON.encode())
                    if str(response.decode().strip(CR)) == ":":
                        _LOGGER.info("Connection open")
                        return True
                    else:
                        _LOGGER.info("Connection established, "
                                     "but wrong response %r.", response)
        except asyncio.TimeoutError:
            _LOGGER.error("Timeout error during connection")
        except SerialException:
            _LOGGER.error("Device not found")
        return self.closed_connection_info()

    def closed_connection_info(self):
        _LOGGER.info("Cannot open serial to Epson")
        return False

    def close(self):
        if self._writer and not self._writer.is_closing():
            self._writer.close()

    async def get_property(self, command, timeout):
        """Get property state from device."""
        response = await self.send_request(
            timeout=timeout,
            command=command+GET_CR)
        if not response:
            return False
        try:
            return response.split('=')[1]
        except KeyError:
            return BUSY
        except IndexError:
            _LOGGER.error("Bad response %s", response)
            return False

    async def send_command(self, command, timeout):
        """Send command to Epson."""
        response = await self.send_request(
            timeout=timeout,
            command=command+CR)
        return response

    async def send_request(self, timeout, command):
        """Send request to Epson over serial."""
        if self._writer is None:
            await self.async_init()
        if self._writer and not self._writer.is_closing() and command:
            try:
                with async_timeout.timeout(timeout):
                    self._writer.write(command.encode())
                    response = await self._reader.readuntil(
                        COLON.encode())
                    response = response[:-1].decode().rstrip(CR)
                    _LOGGER.info("Response from Epson %r", response)
                    if response == ERROR:
                        _LOGGER.error("Error request")
                    else:
                        return response
            except asyncio.TimeoutError:
                _LOGGER.error("Timeout error during sending request")
        return False
